/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#include <string>
#include <thread>
#include <stdexcept>
#include <type_traits>

#include <jni.h>

#include <boost/log/trivial.hpp>

#define REQ_DEF_FILTER // IpAddress
#define REQ_DEF_SERVICE
#include <cdos/defs.h>

#include <cdos/Application.h>
#include <cdos/Configuration.h>

Application g_app;
std::string g_error;

static ServicePtr getService(const std::string& name) {
    try {
        return g_app.getService(name);
    } catch (const std::out_of_range&) {
        g_error = "JNI: Service not found: " + name;
        BOOST_LOG_TRIVIAL(warning) << g_error;

        return ServicePtr();
    }
}

#ifdef __cplusplus
extern "C" {
#endif

//
// Application wrappers
//

// Application nativeGetError
JNIEXPORT jstring JNICALL Java_CounterDoSJ_Application_nativeGetError
(JNIEnv* env, jobject obj)
{
    return env->NewStringUTF(g_error.c_str());
}

// Application nativeStart
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeStart
  (JNIEnv* env, jobject obj, jstring jconfig)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: nativeStart called";

    const char* configStr = env->GetStringUTFChars(jconfig, 0);
    std::string config(configStr);
    env->ReleaseStringUTFChars(jconfig, configStr);

    bool loadStatus = g_app.loadConfigString(config);

    if (!loadStatus) {
        g_error = "Invalid Application configuration";
        return false;
    }

    std::thread(&Application::start, std::ref(g_app)).detach();

    return true;
}

// Application nativeStop
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeStop
  (JNIEnv* env, jobject obj)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: nativeStop called";

    g_app.stop();
    return true;
}

// Application nativeStartService
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeStartService
  (JNIEnv* env, jobject obj, jstring jname, jstring jconfig)
{
    const char* nameStr = env->GetStringUTFChars(jname, 0);
    const char* configStr = env->GetStringUTFChars(jconfig, 0);

    std::string name(nameStr);
    std::string config(configStr);

    env->ReleaseStringUTFChars(jname, nameStr);
    env->ReleaseStringUTFChars(jconfig, configStr);

    Configuration configuration;
    configuration.loadString(config);

    BOOST_LOG_TRIVIAL(trace)
        << "JNI: add service "
        << name;

    try {
        ServicePtr servicePtr(new Service(configuration));
        g_app.addService(name, servicePtr);

        std::thread([=] {
            servicePtr->start();
        }).detach();

        return true;
    } catch (const boost::system::system_error& ex) {
        g_error = std::string("Failed to create service: ") + ex.what();
        BOOST_LOG_TRIVIAL(error) << g_error;

        return false;
    } catch (const std::invalid_argument& ex) {
        g_error = std::string("Failed to create service: ") + ex.what();
        BOOST_LOG_TRIVIAL(error) << g_error;

        return false;
    }
}

// Application nativeStopService
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeStopService
  (JNIEnv* env, jobject obj, jstring jname)
{
    const char* nameStr = env->GetStringUTFChars(jname, 0);
    std::string name(nameStr);
    env->ReleaseStringUTFChars(jname, nameStr);

    BOOST_LOG_TRIVIAL(trace)
        << "JNI: remove service "
        << name;

    ServicePtr servicePtr = getService(name);
    if (!servicePtr) {
        return false;
    }

    servicePtr->stop();
    g_app.removeService(name);

    return true;
}

// Application nativeUpdateCriticalLoad
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeUpdateCriticalLoad
  (JNIEnv* env, jobject obj, jstring jname, bool isCritical)
{
    const char* nameStr = env->GetStringUTFChars(jname, 0);
    std::string name(nameStr);
    env->ReleaseStringUTFChars(jname, nameStr);

    BOOST_LOG_TRIVIAL(trace)
        << "JNI: set critical flag for service "
        << name
        << " to "
        << isCritical;

    ServicePtr servicePtr = getService(name);
    if (!servicePtr) {
        return false;
    }

    servicePtr->setCriticalLoad(isCritical);

    return true;
}

// Application nativeChangeFiltering
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeChangeFiltering
  (JNIEnv* env, jobject obj, jstring jname, int filteringMethod)
{
    const char* nameStr = env->GetStringUTFChars(jname, 0);
    std::string name(nameStr);
    env->ReleaseStringUTFChars(jname, nameStr);

    BOOST_LOG_TRIVIAL(trace)
        << "JNI: change filtering method of service "
        << name
        << " to "
        << filteringMethod
        << " (0: no filter, 1: whitelist, 2: blacklist)";

    ServicePtr servicePtr = getService(name);
    if (!servicePtr) {
        return false;
    }

    FILTER filter;
    switch (filteringMethod) {
    case 0:
        filter = NO_FILTER;
        break;
    case 1:
        filter = WHITELIST;
        break;
    case 2:
        filter = BLACKLIST;
        break;
    default:
        g_error = std::string("Invalid filteringMethod");
        BOOST_LOG_TRIVIAL(warning)
            << g_error << ": " << filteringMethod;
        break;
    }

    servicePtr->setFilter(filter);
    return true;
}

// Application nativeGetServiceMetrics
JNIEXPORT jobject JNICALL Java_CounterDoSJ_Application_nativeGetServiceMetrics
  (JNIEnv* env, jobject obj, jstring jname)
{
    const char* metricsClassName = "CounterDoSJ/ServiceMetrics";
    const char* metricsCtorId = "(JJJJJJ)V";

    const char* nameStr = env->GetStringUTFChars(jname, 0);
    std::string name(nameStr);
    env->ReleaseStringUTFChars(jname, nameStr);

    BOOST_LOG_TRIVIAL(trace)
        << "JNI: get metrics for service "
        << name;

    ServicePtr servicePtr = getService(name);
    if (!servicePtr) {
        return env->NewGlobalRef(NULL);
    }

    const Metrics metrics = servicePtr->getMetrics();

    // create java object
    // TODO cache jclass and jmethodID

    // locate class
    jclass jServiceMetricsClass
        = env->FindClass(metricsClassName);

    if (jServiceMetricsClass == NULL) {
        g_error = std::string("class ") + metricsClassName + " not found";
        BOOST_LOG_TRIVIAL(error) << "JNI: " << g_error;

        return env->NewGlobalRef(NULL);
    }

    // locate constructor
    jmethodID constructor = env->GetMethodID(
        jServiceMetricsClass,
        "<init>",
        metricsCtorId
    );

    if (constructor == NULL) {
        g_error = std::string("method ") + metricsCtorId + " not found";
        BOOST_LOG_TRIVIAL(error) << "JNI: " << g_error;

        return env->NewGlobalRef(NULL);
    }

    BOOST_LOG_TRIVIAL(trace)
        << "metrics: "
        << metrics.totalIncomingRequestCount << "/"
        << metrics.totalOutgoingResponseCount  << "/"
        << metrics.totalDroppedFromQueueCount  << "/"
        << metrics.currentQueueSize  << "/"
        << metrics.totalBlacklistRejection << "/"
        << metrics.totalWhitelistRejection;

    jlong incoming = metrics.totalIncomingRequestCount;
    jlong outgoing = metrics.totalOutgoingResponseCount;
    jlong dropped = metrics.totalDroppedFromQueueCount;
    jlong qsize = metrics.currentQueueSize;
    jlong rejectedBlack = metrics.totalBlacklistRejection;
    jlong rejectedWhite = metrics.totalWhitelistRejection;


    // create object
    return env->NewObject(
        jServiceMetricsClass,
        constructor,
        incoming,
        outgoing,
        dropped,
        qsize,
        rejectedBlack,
        rejectedWhite
    );
}

// jstring -> IpAddress
static IpAddress getIp(JNIEnv* env, jstring jip) {
    const char* ipCStr = env->GetStringUTFChars(jip, 0);
    std::string ipStr(ipCStr);
    env->ReleaseStringUTFChars(jip, ipCStr);

    boost::system::error_code ec;
    IpAddress ip = IpAddress::from_string(ipStr, ec);

    if (ec) {
        g_error = std::string("Invalid IP address: ") + ec.message();
        BOOST_LOG_TRIVIAL(warning) << g_error;
    }

    return ip;
}

// Application nativeBlockIp
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeBlockIp
  (JNIEnv* env, jobject obj, jstring jip)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: block IP address";

    IpAddress ip = getIp(env, jip);
    if (ip.is_unspecified()) {
        return false;
    }

    g_app.blockIp(ip);

    return true;
}

// Application nativeUnblockIp
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeUnblockIp
  (JNIEnv* env, jobject obj, jstring jip)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: unblock IP address";

    IpAddress ip = getIp(env, jip);
    if (ip.is_unspecified()) {
        return false;
    }

    g_app.unblockIp(ip);

    return true;
}

// Application nativeWhitelistIp
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeWhitelistIp
  (JNIEnv* env, jobject obj, jstring jip)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: whitelist IP address";

    IpAddress ip = getIp(env, jip);
    if (ip.is_unspecified()) {
        return false;
    }

    g_app.whitelistIp(ip);

    return true;
}

// Application nativeUnwhitelistIp
JNIEXPORT jboolean JNICALL Java_CounterDoSJ_Application_nativeUnwhitelistIp
  (JNIEnv* env, jobject obj, jstring jip)
{
    BOOST_LOG_TRIVIAL(trace)
        << "JNI: unwhitelist IP address";

    IpAddress ip = getIp(env, jip);
    if (ip.is_unspecified()) {
        return false;
    }

    g_app.unwhitelistIp(ip);

    return true;
}

#ifdef __cplusplus
}
#endif
