/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#include <map>
#include <stdexcept>
#include <thread>

#include <boost/config.hpp> // BOOST_NO_CXX11_HDR_INITIALIZER_LIST
#include <boost/log/core.hpp>
#include <boost/log/trivial.hpp>
#include <boost/log/expressions.hpp>

#include <cdos/Application.h>

Application::Application()
    :_blacklistPtr(new IpFilter()),
     _whitelistPtr(new IpFilter())
{}

bool Application::loadConfig(const std::string& fileName) {
	bool loadStatus = _config.loadFile(fileName);
	if (loadStatus) {
	    bootstrapLogger();
	    bootstrapServices();
	}
	return loadStatus;
}

bool Application::loadConfigString(const std::string& config) {
    bool loadStatus = _config.loadString(config);
    if (loadStatus) {
        bootstrapLogger();
        bootstrapServices();
    }
    return loadStatus;
}

void Application::addService(const std::string& name, ServicePtr servicePtr) {
    // check if service of the same name already exists
    if (_services.find(name) != _services.end()) {
        throw std::invalid_argument("Service already exists with the given name: " + name);
    }

    servicePtr->setBlacklist(_blacklistPtr);
    servicePtr->setWhitelist(_whitelistPtr);
    _services[name] = servicePtr;
}

ServicePtr Application::getService(const std::string& name) {
    return _services.at(name);
}

void Application::removeService(const std::string& name) {
    auto it = _services.find(name);
    _services.erase(it);
}

void Application::start() {
    BOOST_LOG_TRIVIAL(info) << "CDoS Application starts";

    for (auto& serviceEntry : _services) {
        std::thread([=] {
            serviceEntry.second->start();
        }).detach();
    }

    std::unique_lock<std::mutex> waitLock(_mutex);
    _runSignal.wait(waitLock);
}

void Application::stop() {
    BOOST_LOG_TRIVIAL(info) << "CDoS Application stops";

    for (auto& serviceEntry : _services) {
        serviceEntry.second->stop();
    }

    _runSignal.notify_all();
}

void Application::blockIp(IpAddress ip) {
    _blacklistPtr->add(ip);
}

void Application::unblockIp(IpAddress ip) {
    _blacklistPtr->remove(ip);
}

void Application::whitelistIp(IpAddress ip) {
    _whitelistPtr->add(ip);
}

void Application::unwhitelistIp(IpAddress ip) {
    _whitelistPtr->remove(ip);
}

void Application::bootstrapLogger() {
    using boost::log::trivial::severity_level;

    std::string logLevel = _config.get<std::string>("logLevel", "info");

#ifndef BOOST_NO_CXX11_HDR_INITIALIZER_LIST

    std::map<std::string, severity_level> switchMap = {
        {"trace", severity_level::trace},
        {"debug", severity_level::debug},
        {"info", severity_level::info},
        {"warning", severity_level::warning},
        {"error", severity_level::error},
        {"fatal", severity_level::fatal},
    };

    try {
        severity_level severityLevel = switchMap.at(logLevel);

        boost::log::core::get()->set_filter(
            boost::log::trivial::severity >= severityLevel
        );
    } catch (const std::out_of_range& ex) {
        BOOST_LOG_TRIVIAL(warning)
            << "Invalid log level in configuration: "
            << logLevel;
    }

#else
	// MSVC does not support initializer lists (yet?)
	severity_level severityLevel;

	if (logLevel == "trace") {
		severityLevel = severity_level::trace;
	} else if (logLevel == "debug") {
		severityLevel = severity_level::debug;
	} else if (logLevel == "info") {
		severityLevel = severity_level::info;
	} else if (logLevel == "warning") {
		severityLevel = severity_level::warning;
	} else if (logLevel == "error") {
		severityLevel = severity_level::error;
	} else if (logLevel == "fatal") {
		severityLevel = severity_level::fatal;
	} else {
		BOOST_LOG_TRIVIAL(warning)
            << "Invalid log level in configuration: "
            << logLevel;

		severityLevel = severity_level::trace;
	}

	boost::log::core::get()->set_filter(
        boost::log::trivial::severity >= severityLevel
    );

#endif
}

void Application::bootstrapServices() {
    Configuration configs = _config.subtree("services");

    for (auto& it : configs) {
        // create service
        BOOST_LOG_TRIVIAL(info) << "Create service: " << it.first;

        try {
            ServicePtr servicePtr(new Service(it.second));
            addService(it.first, servicePtr);
        } catch (const boost::system::system_error& ex) {
            BOOST_LOG_TRIVIAL(error)
                << "Failed to create service: "
                << ex.what();
        } catch (const std::invalid_argument& ex) {
            BOOST_LOG_TRIVIAL(error)
                << "Failed to create service: "
                << ex.what();
        }
    }
}
