/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef APPLICATION_H_
#define APPLICATION_H_

#include <string>
#include <map>
#include <mutex> // mutex unique_lock
#include <condition_variable>

#define REQ_DEF_FILTER
#define REQ_DEF_SERVICE
#include <cdos/defs.h>

#include <cdos/Configuration.h>
#include <cdos/Filter.hpp>
#include <cdos/Service.h>

/**
 * Represents a running instance of the DoS protection.
 *
 * Manages the running services, configuration and
 * blacklisting. Main entry point of interaction coming
 * from other programs or components.
 */
class Application {
public:
    /**
     * Default constructor
     */
	Application();

	/**
	 * Loads the configuration from the given .json file
	 *
	 * Example configuration:
	 * @code{.json}
	 * {
     *      "logLevel" : "info",
     *      "services" : {
     *          "apache2" : {
     *              "ipAddress" : "127.0.0.1",
     *              "publicPort" : 8090,
     *              "privatePort" : 80,
     *              "maxConcurrentClient" : 3,
     *              "maxQueueSize" : 1500
     *          },
     *          "apache2ssl" : {
     *              "ipAddress" : "127.0.0.1",
     *              "publicPort" : 8091,
     *              "privatePort" : 443,
     *          }
     *      }
     * }
	 * @endcode
	 *
	 * The logLevel key sets the required severity of log
	 * messages to be printed. Available levels:
	 * trace, debug, info, warning, error, fatal.
	 *
	 * @see Application::bootstrapLogger
	 *
	 * The services key contains the configurations
	 * of services to be loaded (but not started) automatically.
	 * For the details of the Service configuration keys,
	 * @see Service::Service. This feature is intended
	 * to be used in a standalone scenario, the OSGi
	 * wrapper has other ways to configure services.
	 *
	 * @param fileName full path and name of the config file
	 * @return true if config loaded, false on failure
	 */
	bool loadConfig(const std::string& fileName);

	/**
	 * Loads config from string
	 *
	 * @see loadConfig
	 *
	 * @param JSON config string
	 * @return true if config loaded, false on failure
	 */
	bool loadConfigString(const std::string& config);

	/**
	 * Adds the given service to the service container,
	 * identified by the given name. Sets the blacklist filter.
	 *
	 * @param name Service name (identifier)
	 * @param servicePtr Service to add
	 *
	 * @throw std::invalid_argument if a service is already registered
	 * with the given name
	 */
	void addService(const std::string& name, ServicePtr servicePtr);

	/**
	 * Gets the service with the given name (identifier)
	 *
	 * @param name Name of the service to retrive
	 * @return Service with the given name
	 *
	 * @throw std::out_of_range if no such service is present
	 */
	ServicePtr getService(const std::string& name);

	/**
	 * Removes the service with the given name
	 * from the service container.
	 *
	 * This operation does not affect the run
	 * status of any service.
	 *
	 * @param name Name of the service to remove
	 */
	void removeService(const std::string& name);

	/**
	 * Convenience method to start all of the
	 * registered services in a new thread,
	 * and block the current thread until
	 * stop() is called.
	 *
	 * @see Application::stop()
	 */
	void start();

	/**
	 * Convenience method to stop all of the
     * registered services.
	 */
	void stop();

	/**
	 * Blacklists the given IP address
	 *
	 * @param ip IP address to blacklist
	 */
	void blockIp(IpAddress ip);

	/**
	 * Removes the given IP address from the blacklist
	 *
	 * @param ip IP address to unblock
	 */
	void unblockIp(IpAddress ip);

	/**
	 * Whitelists the given IP address
	 *
	 * @param ip IP address to whitelist
	 */
	void whitelistIp(IpAddress ip);

	/**
	 * Removes the given IP address from the whitelist
	 *
	 * @param ip IP address to remove
	 */
	void unwhitelistIp(IpAddress ip);

private:
	/**
	 * Sets up logger by configuration
	 */
	void bootstrapLogger();

	/**
	 * Sets up services by configuration
	 */
	void bootstrapServices();

	IpFilterPtr _blacklistPtr;
	IpFilterPtr _whitelistPtr;
	Configuration _config;
	std::map<std::string, ServicePtr> _services;
	std::mutex _mutex;
	std::condition_variable _runSignal;
};

#endif /* APPLICATION_H_ */
