/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef BRIDGE_H_
#define BRIDGE_H_

#include <utility> // move
#include <thread>
#include <chrono>

#include <boost/asio.hpp>

#include <boost/log/trivial.hpp>

/**
 * Connects two sockets
 *
 * Reads from the socket on the left, and writes
 * what has been read to the socket on the right,
 * and vice-versa.
 */
template <typename Protocol, typename Left, typename Right>
class Bridge {
public:
    /**
     * Creates bridge, takes ownership of the given sockets
     *
     * @param left Left socket
     * @param right Right socket
     */
    Bridge(Left& left, Right& right)
        :_left(std::move(left)),
         _right(std::move(right))
    {}

    /**
     * Starts reading and writing. Blocks until
     * EOF is reached in both of the sockets.
     */
    void run() {
        Worker<Left, Right> lr(_left, _right);
        Worker<Right, Left> rl(_right, _left);

        std::thread lrThread(
            &Worker<Left, Right>::run, std::ref(lr));

        rl.run();

        lrThread.join();
    }

private:
    template <typename WLeft, typename WRight>
    class Worker {
    public:
        Worker(WLeft& left, WRight& right)
            :_left(left),
             _right(right),
             _leftPort(0),
             _rightPort(0)
        {
            try {
                _leftPort = _left.remote_endpoint().port();
            } catch (const boost::system::system_error& ex) {
                BOOST_LOG_TRIVIAL(warning)
                    << "Bridge: Failed to get remote endpoint of left socket: "
                    << ex.what();
            }

            try {
                _rightPort = _right.remote_endpoint().port();
            } catch (const boost::system::system_error& ex) {
                BOOST_LOG_TRIVIAL(warning)
                    << "Bridge: Failed to get remote endpoint of right socket: "
                    << ex.what();
            }
        }

        void run() {
            int emptyReceiveCount = 0;  // see below

            while (true) {
                boost::system::error_code status;
                BOOST_LOG_TRIVIAL(trace)
                    << "[" << _leftPort << "] -R-";

                size_t dataSize = _left.receive(
                    boost::asio::buffer(_data, maxLength),
                    0, // socket_base::message_flags
                    status
                );

                BOOST_LOG_TRIVIAL(trace)
                    << "[" << _leftPort << "] --- " << dataSize << " bytes";

                if (status == boost::asio::error::eof) {
                    BOOST_LOG_TRIVIAL(trace)
                        << "[" << _leftPort << "] --- EOF --- [" << _rightPort << "]";

                    boost::system::error_code shutdownStatus;
                    _right.shutdown(Protocol::base::socket::shutdown_receive, shutdownStatus);

                    if (shutdownStatus) {
                        BOOST_LOG_TRIVIAL(debug)
                            << "Socket was already closed: "
                            << shutdownStatus;
                    }

                    return; // connection closed
                } else if (dataSize == 0 && status == 0) {
                    // experiment shows, while using UDP
                    // it's possible to read 0 bytes without error;
                    // however, an early close might lead to a
                    // race condition, which would cause a
                    // message loss.
                    ++emptyReceiveCount;
                    if (emptyReceiveCount > Protocol::maxEmptyResponse) {
                        BOOST_LOG_TRIVIAL(trace)
                            << "[" << _leftPort << "] --- END --- [" << _rightPort << "]";

                        return;
                    }

                    // sleep a bit and give a chance to
                    // the private service to catch up.
                    std::this_thread::sleep_for(
                        std::chrono::milliseconds(10 * emptyReceiveCount)
                    );

                } else if (status != 0) {
                    BOOST_LOG_TRIVIAL(warning)
                        << "Failed to read socket, status code: "
                        << status;
                    return;
                } else if (dataSize > 0) {
                    // this condition is required because
                    // of the empty response case of UDP

                    // allow no more empty reads
                    emptyReceiveCount = Protocol::maxEmptyResponse;

                    BOOST_LOG_TRIVIAL(trace)
                        << "[" << _leftPort << "] --- " << dataSize << " bytes -W- [" << _rightPort << "]";

                    _right.send(
                        boost::asio::buffer(_data, dataSize),
                        0, // socket_base::message_flags
                        status
                    );

                    if (status != 0) {
                        BOOST_LOG_TRIVIAL(warning)
                            << "Failed to write socket, status code: "
                            << status;
                        return;
                    }
                }

            }
        }

    private:
        WLeft& _left;
        WRight& _right;

        unsigned short _leftPort;
        unsigned short _rightPort;

        enum { maxLength = Protocol::bufferLength };
        char _data[maxLength];
    };

    Left _left;
    Right _right;
};

#endif /* BRIDGE_H_ */
