/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef CONFIGURATION_H_
#define CONFIGURATION_H_

#include <string>
#include <stdexcept>

#include <boost/property_tree/ptree.hpp>

/**
 * Represents configuration
 */
class Configuration {
public:
    typedef boost::property_tree::ptree::iterator iterator;

    /**
     * Default constructor, empty configuration.
     */
	Configuration() {}

	/**
	 * Creates a new Configuration object from
	 * the underlying data structure.
	 *
	 * This constructor is present to ease the
	 * work with the iterator.
	 *
	 * @param data Underlying data structure
	 */
	Configuration(const boost::property_tree::ptree& data);

	/**
	 * Loads configuration from the given .json file
	 *
	 * @note Clears existing contents. In case of error the
     * Configuration is unmodified.
     *
	 * @param fileName full path and name of the config file
	 * @return true if config loaded, false on failure
	 */
	bool loadFile(const std::string& fileName);

	/**
	 * Loads configuration from the given json string
	 *
	 * @note Clears existing contents. In case of error the
     * Configuration is unmodified.
     *
     * @param config JSON string
     * @return true if config loaded, false on failure
	 */
	bool loadString(const std::string& config);

	/**
	 * Gets a configuration object representing
	 * a subtree of this configuration
	 * identified by the given key.
	 *
	 * @param key path to the required subtree,
	 * e.g: top.middle.key
	 *
	 * @return The requested subtree, empty
	 * if key not found
	 */
	Configuration subtree(const char* key);

	/**
	 * Gets the value of the given key
	 *
	 * @param key Key
	 * @return Value of key
	 *
	 * @throw std::logic_error if can't convert data
	 * to the requested Value type
	 */
	template <typename Value>
	Value get(const char* key) const {
	    try {
            return _data.get<Value>(key);
	    } catch (const boost::property_tree::ptree_bad_data& ex) {
	        throw std::logic_error(ex.what());
	    }
	}

	/**
	 * Gets the value of the given key,
	 * or defaultValue, if key not found or
	 * conversion fails.
	 *
	 * @param key Key
	 * @param defaultValue Default value
	 *
	 * @return Value of key or defaultValue if key
	 * not found or the conversion fails
	 */
	template <typename Value>
	Value get(const char* key, Value defaultValue) const {
		return _data.get<Value>(key, defaultValue);
	}

	/**
	 * Gets the count of the top level keys
	 *
	 * @return Count of the top level keys
	 */
	inline size_t size() {
	    return _data.size();
	}

	/**
	 * @return Iterator pointing to the
	 * first top level key
	 */
	inline iterator begin() {
	    return _data.begin();
	}

	/**
	 * @return Iterator pointing after
	 * the last top level key
	 */
	inline iterator end() {
        return _data.end();
    }

private:
	boost::property_tree::ptree _data;

};

#endif /* CONFIGURATION_H_ */
