/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef FILTER_HPP_
#define FILTER_HPP_

#include <vector>
#include <algorithm> // remove, find
#include <mutex> // mutex lock_guard

/**
 * Generic template of the following features:
 *  - Container of items
 *  -- add
 *  -- remove
 *  -- has
 *  - Thread safe
 *
 * @tparam T Type of items to store
 */
template <typename T>
class Filter {
public:
    /**
     * Adds an item to the container
     *
     * @param item Item to add
     */
    void add(const T& item) {
        std::lock_guard<std::mutex> guard(_mutex);
        _filtered.push_back(item);
    }

    /**
     * Removes all items from the container
     * which are equal to the given item
     *
     * @param item Item to remove
     */
    void remove(const T& item) {
        std::lock_guard<std::mutex> guard(_mutex);

        _filtered.erase(
            std::remove(
                _filtered.begin(),
                _filtered.end(),
                item),
            _filtered.end()
        );
    }

    /**
     * Checks existence of the given item
     * in the container
     *
     * @param item Item to check
     * @return true if the container contains
     * the item, false otherwise
     */
    bool has(const T& item) const {
        std::lock_guard<std::mutex> guard(_mutex);

        return
            std::find(_filtered.begin(), _filtered.end(), item)
        !=  _filtered.end();
    }

private:
    std::vector<T> _filtered;
    mutable std::mutex _mutex;
};

#endif /* FILTER_HPP_ */
