/*
* Copyright (c) 2013 SEARCH-LAB Ltd.
* 
*/

#include <boost/log/trivial.hpp>

#include <cdos/Protocol.h>


TCP::acceptor::acceptor(
    boost::asio::io_service& ioService,
    tcp::endpoint endpoint
)
    :_acceptor(ioService, endpoint)
{}

void TCP::acceptor::accept(socket& socket, const AcceptCb& callback) {
    _acceptor.async_accept(
        socket,
        [&](boost::system::error_code errorCode) {
            if (!errorCode) {
                // forward
                callback(std::move(socket));
            }

            accept(socket, callback);
        }
    );
}

void TCP::acceptor::close(boost::system::error_code& ec) {
    _acceptor.close(ec);
}


UDP::request::request()
    :_data(new char[maxLength]),
     _size(0)
{}

UDP::request::request(request&& rhs)
    :_remoteEndpoint(rhs._remoteEndpoint),
     _data(std::move(rhs._data)),
     _size(rhs._size)
{
    rhs._data.reset(new char[maxLength]);
    rhs._size = 0;
}

UDP::request::request(const boost::asio::io_service&)
    :_size(0)
{}

void UDP::request::operator=(request&& rhs) {
    _remoteEndpoint = rhs._remoteEndpoint;
    _data = std::move(rhs._data);
    _size = rhs._size;

    rhs._data.reset(new char[maxLength]);
    rhs._size = 0;
}

void UDP::request::connect(base::endpoint endpoint) {
    _remoteEndpoint = endpoint;
}

UDP::base::endpoint& UDP::request::remote_endpoint() {
    return _remoteEndpoint;
}

boost::asio::mutable_buffers_1 UDP::request::buffer() {
    return boost::asio::buffer(
        _data.get(),
        (_size) ? _size : maxLength
    );
}

void UDP::request::setSize(std::size_t size) {
    _size = size;
}

std::size_t UDP::request::receive(
    boost::asio::mutable_buffers_1 otherBuffer,
    int, // socket_base::message_flags
    boost::system::error_code& ec
) {
    if (is_open()) {
        std::size_t received = boost::asio::buffer_copy(otherBuffer, buffer());
        _data.reset();

        return received;
    } else {
        ec = boost::asio::error::eof;
        return 0;
    }
}

void UDP::request::send(
    boost::asio::mutable_buffers_1 otherBuffer,
    boost::asio::socket_base::message_flags flags,
    boost::system::error_code& ec
) {
    boost::asio::io_service ioService;
    base::socket privateSocket(ioService);
    privateSocket.connect(_remoteEndpoint, ec);

    if (ec) {
        return;
    }

    privateSocket.send(otherBuffer, flags, ec);
}

bool UDP::request::is_open() const {
    return (_data) ? true : false;
}

UDP::acceptor::acceptor(
    boost::asio::io_service& ioService,
    udp::endpoint endpoint
)
    :_endpoint(endpoint)
{}

void UDP::acceptor::accept(socket& socket, const AcceptCb& callback) {
    boost::system::error_code ec;
    socket.open(udp::v4(), ec);

    if (ec) {
        BOOST_LOG_TRIVIAL(error)
            << "Failed to open UDP socket: "
            << ec;
        return;
    }

    socket.bind(_endpoint, ec);

    if (ec) {
        BOOST_LOG_TRIVIAL(error)
            << "Failed to bind UDP socket to port "
            << _endpoint.port();
        return;
    }

    doAccept(socket, callback);
}

void UDP::acceptor::doAccept(socket& socket, const AcceptCb& callback) {
    socket.async_receive_from(
        _request.buffer(),
        _request.remote_endpoint(),
        [this, &socket, &callback](boost::system::error_code ec, std::size_t size)
        {
            if (ec) {
                BOOST_LOG_TRIVIAL(warning)
                    << "Failed to read UDP socket: "
                    << ec;
            } else if (size > 0) {
                _request.setSize(size);
                callback(std::move(_request));
            }

            doAccept(socket, callback);
        }
    );
}
