/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef QUEUE_HPP_
#define QUEUE_HPP_

#include <deque>
#include <utility> // move
#include <mutex> // mutex unique_lock
#include <stdexcept> // length_error
#include <condition_variable>
#include <algorithm>

#include <boost/log/trivial.hpp>

/**
 * Threadsafe queue of limited amount of
 * movable objects
 *
 * Original implementation is at:
 * http://juanchopanzacpp.wordpress.com/2013/02/26/concurrent-queue-c11/
 *
 * @tparam T type of items to enqueue
 */
template <typename T>
class Queue {
public:
    /**
     * Create an empty queue
     *
     * @param maxSize maximum size of the queue
     */
    Queue(unsigned maxSize)
        :_maxSize(maxSize),
         _dropCount(0)
    {}

    /**
     * Add an object to the end of the queue by move.
     *
     * If the queue exceeds its size limit,
     * the first half of the queue will be
     * dropped.
     *
     * @param item to add
     */
    void push(T&& item) {
        {
            std::unique_lock<std::mutex> lock(_mutex);
            _queue.push_back(std::move(item));

            if (_queue.size() > _maxSize) {
                BOOST_LOG_TRIVIAL(info)
                    << "Queue length is over allowed size "
                    << _maxSize << ". Initiate trimming";

                _dropCount += _queue.size()/2;

                _queue.erase(
                    _queue.begin(),
                    _queue.begin() + _queue.size()/2
                );
            }
        }
        _signal.notify_one();
    }

    /**
     * Removes an item from the beginning
     * of the queue and returns it.
     *
     * Blocks if the queue is empty
     *
     * @return Item which was found on the beginning
     */
    T pop() {
        std::unique_lock<std::mutex> lock(_mutex);
        while (_queue.empty()) {
            _signal.wait(lock);
        }

        T item = std::move(_queue.front());
        _queue.pop_front();
        return item;
    }

    /**
     * @return Size of the queue
     */
    size_t size() const {
        std::unique_lock<std::mutex> lock(_mutex);
        return _queue.size();
    }

    /**
     * @return Count of dropped items
     * because of size limit
     */
    size_t dropCount() const {
        std::unique_lock<std::mutex> lock(_mutex);
        return _dropCount;
    }

private:
    unsigned _maxSize;
    unsigned _dropCount;
    mutable std::mutex _mutex;
    std::condition_variable _signal;
    std::deque<T> _queue;
};

#endif /* QUEUE_HPP_ */
