/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef SERVER_H_
#define SERVER_H_

#include <functional> // function
#include <utility> // move
#include <thread>

#include <boost/asio.hpp>
#include <boost/log/trivial.hpp>

/**
 * Generic network server listening on a port
 * and accepting connections.
 */
template <typename Protocol>
class Server {
public:
    /**
     * Create Server instance
     *
     * @param listenPort port to listen on
     */
    Server(unsigned listenPort)
        :_listenPort(listenPort),
         _acceptor(
             _ioService,
             typename Protocol::base::endpoint(
                 Protocol::base::v4(),
                 listenPort
             )
         ),
         _socket(_ioService)
    {}

    /**
     * Starts the server
     *
     * Accepts each incoming connection
     * and calls the given callback
     * passing the socket.
     */
    void start(const typename Protocol::AcceptCb& acceptCb) {
        std::thread acceptorThread([&]{
            _acceptor.accept(_socket, acceptCb);
            _ioService.run();
        });

        acceptorThread.join();
    }

    /**
     * Stops the server
     *
     * No more connection will be accepted.
     */
    void stop() {
        BOOST_LOG_TRIVIAL(trace)
            << "Stopping Server on port "
            << _listenPort;

        _ioService.stop();

        boost::system::error_code ec;
        _acceptor.close(ec);

        if (ec) {
            BOOST_LOG_TRIVIAL(warning)
                << "Failed to close Server acceptor: "
                << ec;
        }

        _socket.close(ec);

        if (ec) {
            BOOST_LOG_TRIVIAL(warning)
                << "Failed to close Server socket: "
                << ec;
        }

        BOOST_LOG_TRIVIAL(trace)
            << "Server stopped";
    }

private:
    unsigned _listenPort;
    boost::asio::io_service _ioService;
    typename Protocol::acceptor _acceptor;
    typename Protocol::socket _socket;
};


#endif /* SERVER_H_ */
