/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef SERVICE_H_
#define SERVICE_H_

#include <mutex>
#include <condition_variable>
#include <memory>
#include <atomic>

#include <boost/asio.hpp>

#define REQ_DEF_FILTER
#define REQ_DEF_METRICS
#include <cdos/defs.h>

#include <cdos/Configuration.h>
#include <cdos/ServiceImpl.hpp>

/**
 *  Represents a DoS protecting tunnel
 *  which forwards connections from a public port
 *  to a private one while guarding the private
 *  service against malicious attacks.
 *
 *  Overview:
 *  // listen on public port
 *  // on incoming request
 *  // if on blacklist
 *      // drop
 *  // else
 *      // put request into the wait queue
 *
 *  // in the same time, serve queue items
 */
class Service {
public:
    /**
     * Create a service from the given configuration
     *
     * Required config keys:
     *  - ipAddress: IP addr. of the private service
     *  - publicPort
     *  - privatePort
     *
     * Optional config keys:
     *  - maxConcurrentClient: queue consumer thread count
     *  - maxQueueSize
     *  - protocol: tcp or udp, default is tcp
     *
     * @param config Configuration
     *
     * @throw boost::system::error_code on invalid ip
     * @throw std::logic_error if required config key
     * is missing or in wrong format
     * @throw std::logic_error if invalid protocol specified
     */
    Service(const Configuration& config);

    /**
     * Sets filtering method
     */
    void setFilter(FILTER filter);

    /**
     * Sets blacklist instance
     */
    void setBlacklist(const IpFilterPtr& blacklist);

    /**
     * Sets whitelist instance
     */
    void setWhitelist(const IpFilterPtr& whitelist);

    /**
     * Instructs the protection about the
     * system load of the private service.
     *
     * If the load is critical, serving of
     * queued items stops.
     *
     * @param isCritical true if the system load
     * is critical, false otherwise
     */
    void setCriticalLoad(bool isCritical);

    /**
     * Starts the Service
     *
     * @throw boost::system::system_error if the server
     * fails to start
     */
    void start();

    /**
     * Stops the Service
     *
     * Once it's stopped, it can't be restarted.
     */
    void stop();

    /**
     * Get current metrics of the Service
     *
     * @return actual metrics
     */
    const Metrics getMetrics() const;

private:
    enum STATUS {
        CREATED,
        RUNNING,
        STOPPED
    };

    std::atomic<STATUS> _status;
    unsigned _publicPort;
    unsigned _privatePort;

    typedef std::shared_ptr<ServiceImplBase> ServiceImplPtr;
    ServiceImplPtr _impl;
};

#endif /* SERVICE_H_ */
