/*
 * Copyright (c) 2013 SEARCH-LAB Ltd.
 * 
 */

#ifndef SERVICEIMPL_HPP_
#define SERVICEIMPL_HPP_

#include <atomic>
#include <thread>
#include <utility> // move
#include <functional>

#include <boost/asio.hpp>

#define REQ_DEF_FILTER
#define REQ_DEF_METRICS
#include <cdos/defs.h>
#include <cdos/Configuration.h>
#include <cdos/Protocol.h>
#include <cdos/Server.hpp>
#include <cdos/Queue.hpp>
#include <cdos/Bridge.hpp>

class ServiceImplBase {
public:
    virtual ~ServiceImplBase() {}

    virtual void setFilter(FILTER filter) = 0;
    virtual void setBlacklist(const IpFilterPtr& blacklist) = 0;
    virtual void setWhitelist(const IpFilterPtr& whitelist) = 0;

    virtual void startServer() = 0;
    virtual void stopServer() = 0;

    virtual void startProcessors() = 0;
    virtual void stopProcessors() = 0;

    virtual void setCriticalLoad(bool isCritical) = 0;

    virtual const Metrics getMetrics() = 0;
};

template <typename Protocol>
class ServiceImpl
    : public ServiceImplBase,
      public std::enable_shared_from_this<ServiceImplBase>
{
public:
    ServiceImpl(const Configuration& config)
        :_ipAddress(IpAddress::from_string(config.get<std::string>("ipAddress"))),
         _publicPort(config.get<unsigned>("publicPort")),
         _privatePort(config.get<unsigned>("privatePort")),
         _maxConcurrentClient(config.get<unsigned>("maxConcurrentClient", 1)),
         _maxQueueSize(config.get<unsigned>("maxQueueSize", 1000)),
         _incomingServer(_publicPort),
         _queue(_maxQueueSize),
         _currentFilter(BLACKLIST),
         _blacklistPtr(new IpFilter()),
         _whitelistPtr(new IpFilter()),
         _isCriticalLoad(false),
         _isProcessing(false),
         _runningBridgeCount(0)
    {}

    ~ServiceImpl() {};

    void setFilter(FILTER filter) {
        _currentFilter = filter;
    }

    void setBlacklist(const IpFilterPtr& blacklist) {
        _blacklistPtr = blacklist;
    }

    void setWhitelist(const IpFilterPtr& whitelist) {
        _whitelistPtr = whitelist;
    }

    void startServer() {
		// MSVC v110 requires the following typedef (incorrectly)
		typedef typename Protocol::base::endpoint Endpoint;

        _incomingServer.start([this](typename Protocol::request request) {
            // incoming request
            _metrics.incomingRequest();

            // get client ip
            IpAddress remoteIp;
            try {
                Endpoint remoteEnd = request.remote_endpoint();
                remoteIp = remoteEnd.address();
            } catch (const boost::system::system_error& ex) {
                BOOST_LOG_TRIVIAL(warning)
                    << "Failed to get remote end ip, drop socket. "
                    << ex.what();

                return;
            }

            // filtering
            if (isAllowedByFilter(remoteIp) == false) {
                request.close();
                return;
            }

            // still here? add socket to queue
            BOOST_LOG_TRIVIAL(trace)
                << "Add socket to queue";
            _queue.push(std::move(request));
        });
    }

    void stopServer() {
        waitForNoBridge();
        _incomingServer.stop();
    }

    void startProcessors() {
        _isProcessing = true;

        auto self(this->shared_from_this());
        for (unsigned i = 0; i < _maxConcurrentClient; ++i) {
            std::thread processorThread([this, self] {
                process();
            });
            processorThread.detach();
        }
    }

    void stopProcessors() {
        _isProcessing = false;

        // inject closed sockets into the queue
        // to wake up waiting threads
        for (unsigned i = 0; i < _maxConcurrentClient; ++i) {
            typename Protocol::request closedSocket(_dummyIoService);
            _queue.push(std::move(closedSocket));
        }
    }

    void setCriticalLoad(bool isCritical) {
        _isCriticalLoad = isCritical;

        if (!_isCriticalLoad) {
            _loadSignal.notify_all();
        }
    }

    const Metrics getMetrics() {
        return _metrics.snapshot(*this);
    }

private:
    typedef ServiceImpl<Protocol> This;

    class MetricsManager {
    public:
        MetricsManager()
            :_totalIncomingRequestCount(0),
             _totalOutgoingResponseCount(0),
             _totalBlacklistRejection(0),
             _totalWhitelistRejection(0)
        {}

        inline void incomingRequest() {
            ++_totalIncomingRequestCount;
        }

        inline void outgoingResponse() {
            ++_totalOutgoingResponseCount;
        }

        inline void rejectByBlacklist() {
            ++_totalBlacklistRejection;
        }
        inline void rejectByWhitelist() {
            ++_totalWhitelistRejection;
        }

        const Metrics snapshot(const This& service) const {
            Metrics metrics;

            metrics.totalIncomingRequestCount = _totalIncomingRequestCount;
            metrics.totalOutgoingResponseCount = _totalOutgoingResponseCount;
            metrics.totalDroppedFromQueueCount = service._queue.dropCount();
            metrics.currentQueueSize = service._queue.size();
            metrics.totalBlacklistRejection = _totalBlacklistRejection;
            metrics.totalWhitelistRejection = _totalWhitelistRejection;

            return metrics;
        }

    private:
        std::atomic<unsigned> _totalIncomingRequestCount;
        std::atomic<unsigned> _totalOutgoingResponseCount;
        std::atomic<unsigned> _totalBlacklistRejection;
        std::atomic<unsigned> _totalWhitelistRejection;
    };

    bool isAllowedByFilter(IpAddress ip) {
        bool isAllowed = false;

        switch (_currentFilter) {
        case NO_FILTER:
            isAllowed = true;
            break;
        case WHITELIST:
            isAllowed = _whitelistPtr->has(ip);
            if (!isAllowed) {
                _metrics.rejectByWhitelist();

                BOOST_LOG_TRIVIAL(info)
                    << "Drop socket of not whitelisted ip: " << ip;
            }
            break;
        case BLACKLIST:
            isAllowed = (_blacklistPtr->has(ip) == false);
            if (!isAllowed) {
                _metrics.rejectByBlacklist();

                BOOST_LOG_TRIVIAL(info)
                    << "Drop socket of blacklisted ip: " << ip;
            }
            break;
        }

        return isAllowed;
    }

    void waitForNoBridge() {
        std::mutex mutex;
        std::unique_lock<std::mutex> lock(mutex);
        _noBridgeSignal.wait(lock, [this] {
            bool canContinue = (_runningBridgeCount == 0);

            if (!canContinue) {
                BOOST_LOG_TRIVIAL(trace)
                    << "Waiting for "
                    << _runningBridgeCount
                    << " bridges";
            }

            return canContinue;
        });
    }

    void waitForNonCriticalLoad() {
        std::unique_lock<std::mutex> lock(_loadMutex);
        _loadSignal.wait(lock, [this]{ return !_isCriticalLoad; });
    }

    void process() {
        while (_isProcessing) {
            BOOST_LOG_TRIVIAL(trace)
                << "Processor thread is about to get next socket";
            // get next socket
            // pop blocks if the queue is empty
            typename Protocol::request publicSocket = _queue.pop();

            BOOST_LOG_TRIVIAL(trace)
                << "Processor thread picked up socket";

            // wait while load is not critical
            if (_isCriticalLoad) {
                BOOST_LOG_TRIVIAL(trace)
                    << "Processor thread waits for noncritical load";
                waitForNonCriticalLoad();
            }

            // if load changes frequently
            // it might be critical even here.
            // however, assuming the relatively
            // small number of processor threads
            // and WTRAC hysteresis, it's not an issue.

            // are you still there?
            if (publicSocket.is_open() == false) {
                continue; // goto process loop
            }

            try {
                typename Protocol::socket privateSocket(_dummyIoService);
                privateSocket.connect(
                    typename Protocol::base::endpoint(_ipAddress, _privatePort)
                );

                // let's bridge
                typedef Bridge<
                    Protocol,
                    typename Protocol::request,
                    typename Protocol::socket
                > BridgePRS;

                std::shared_ptr<BridgePRS> bridgePtr(
                    new BridgePRS(publicSocket, privateSocket)
                );

                _metrics.outgoingResponse();

                std::thread([bridgePtr, this] () mutable {
                    ++_runningBridgeCount;

                    bridgePtr->run(); // blocks
                    bridgePtr.reset();

                    --_runningBridgeCount;
                    if (_runningBridgeCount == 0) {
                        _noBridgeSignal.notify_one();
                    }

                    BOOST_LOG_TRIVIAL(trace)
                        << "Tunnelling completed";
                }).detach();

            } catch (const boost::system::system_error& ex) {
                BOOST_LOG_TRIVIAL(warning)
                    << "Failed to connect private service "
                    << _ipAddress << ":" << _privatePort
                    << "/" << ex.what() << "/";
            }
        }

        BOOST_LOG_TRIVIAL(trace)
            << "Processor thread terminates";
    }

    IpAddress _ipAddress;
    unsigned _publicPort;
    unsigned _privatePort;
    unsigned _maxConcurrentClient;
    unsigned _maxQueueSize;

    Server<Protocol> _incomingServer;
    Queue<typename Protocol::request> _queue;

    FILTER _currentFilter;
    IpFilterPtr _blacklistPtr;
    IpFilterPtr _whitelistPtr;

    std::atomic<bool> _isCriticalLoad;
    std::mutex _loadMutex;
    std::condition_variable _loadSignal;

    bool _isProcessing;

    std::atomic<unsigned> _runningBridgeCount;
    std::condition_variable _noBridgeSignal;

    MetricsManager _metrics;

    boost::asio::io_service _dummyIoService;
};

#endif /* SERVICEIMPL_HPP_ */
