package CounterDoSJTest.test;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import CounterDoSJ.Tunnel;
import CounterDoSJ.TunnelFactory;
import CounterDoSJTest.network.client.Client;
import CounterDoSJTest.network.client.TCPClient;
import CounterDoSJTest.network.client.UDPClient;
import CounterDoSJTest.network.server.Server;
import CounterDoSJTest.network.server.TCPServer;
import CounterDoSJTest.network.server.UDPServer;

public class Framework {
    private static int tcpServerPort = 12345;
    private static int udpServerPort = 12345;
    
    private static int tcpTunnelPort = 12346;
    private static int udpTunnelPort = 12346;
    
    private static InetAddress localhost;
    
    private static InetAddress clientAHost;
    private static InetAddress clientBHost;
    
    static {
        try {
            // don't use InetAddress.getLocalHost();,
            // which returns 127.0.1.1 while
            // DatagramSocket.send uses 127.0.0.1
            localhost = InetAddress.getByName("127.0.0.1");
            
            clientAHost = InetAddress.getByName("127.0.0.2");
            clientBHost = InetAddress.getByName("127.0.0.3");
        } catch (UnknownHostException ex) {
            System.err.println("Failed to get host address");
            ex.printStackTrace();
        }
    }
    
    public static void setTunnelFactory(TunnelFactory tf) {
        _tunnelFactory = tf;
    }
    
    private static synchronized int getNextTunnelIndex() {
        return _tunnelIndex++;
    }
    
    public boolean runAll() {
        System.out.println("Running CDoSJ tests");
        System.out.println("Localhost is: " + localhost.getHostAddress());
        
        boolean status = true;
        
        status &= runTCP();
        
        if (status || !_stopOnFailure) {
            status &= runUDP();
        }
        
        if (status) {
            System.out.println("SUCCESS");
        } else {
            System.out.println("FAILURE");
        }
        
        return status;
    }
    
    private boolean runTCP() {
        try {
            System.out.println("Running TCP tests");
            
            Server server = new TCPServer(tcpServerPort);
            server.listen();
            
            Client directClient = new TCPClient(clientAHost, localhost, tcpServerPort);
            Client tunnelClient = new TCPClient(clientAHost, localhost, tcpTunnelPort);
            Client tunnelClientB = new TCPClient(clientBHost, localhost, tcpTunnelPort);
            
            Tunnel tunnel = createTCPTunnel(tcpTunnelPort, tcpServerPort);
            tunnel.start();
            
            List<Test> tests = new ArrayList<Test>();
            tests.add(createSendReceiveTest(directClient));
            tests.add(createSendReceiveTest(tunnelClient));
            tests.add(createCriticalLoadTest(tunnelClient, tunnel));
            tests.add(createBlacklistTest(tunnelClient, tunnelClientB, tunnel, _tunnelFactory));
            tests.add(createWhitelistTest(tunnelClient, tunnelClientB, tunnel, _tunnelFactory));
            
            boolean status = true;
            
            for (Test test : tests) {
                status &= runTest(test);
                
                if (!status && _stopOnFailure) {
                    break;
                }
            }
            
            tunnel.stop();
            server.stop();
            
            return status;
        } catch (IOException ex) {
            ex.printStackTrace();
            return false;
        }
    }
    
    private boolean runUDP() {
        try {
            System.out.println("Running UDP tests");
            
            Server server = new UDPServer(udpServerPort);
            server.listen();
            
            Client directClient = new UDPClient(clientAHost, localhost, udpServerPort);
            Client tunnelClient = new UDPClient(clientAHost, localhost, udpTunnelPort);
            Client tunnelClientB = new UDPClient(clientBHost, localhost, udpTunnelPort);
            
            Tunnel tunnel = createUDPTunnel(udpTunnelPort, udpServerPort);
            tunnel.start();
            
            List<Test> tests = new ArrayList<Test>();
            tests.add(createSendReceiveTest(directClient));
            tests.add(createSendReceiveTest(tunnelClient));
            tests.add(createCriticalLoadTest(tunnelClient, tunnel));
            tests.add(createBlacklistTest(tunnelClient, tunnelClientB, tunnel, _tunnelFactory));
            tests.add(createWhitelistTest(tunnelClient, tunnelClientB, tunnel, _tunnelFactory));
            
            boolean status = true;
            
            for (Test test : tests) {
                status &= runTest(test);
                
                if (!status && _stopOnFailure) {
                    break;
                }
            }
            
            tunnel.stop();
            server.stop();
            
            return status;
        } catch (IOException ex) {
            ex.printStackTrace();
            return false;
        }
    }
    
    private boolean runTest(Test test) {
        System.out.println("Run test: " + test.getClass().getSimpleName());
        return test.run();
    }
    
    private Tunnel createTCPTunnel(int publicPort, int privatePort) {
        return _tunnelFactory.createTunnel(
            "testTunnel_" + getNextTunnelIndex(), 
            localhost, publicPort, privatePort
        );
    }
    
    private Tunnel createUDPTunnel(int publicPort, int privatePort) {
        return _tunnelFactory.createTunnel(
            "testTunnel_" + getNextTunnelIndex(), 
            localhost, publicPort, privatePort, Tunnel.PROTOCOL.UDP
        );
    }
    
    private Test createSendReceiveTest(Client client) {
        return new SendReceiveTest(client);
    }
    
    private Test createCriticalLoadTest(Client client, Tunnel tunnel) {
        return new CriticalLoadTest(client, tunnel);
    }
    
    private Test createBlacklistTest(Client clientA, Client clientB, Tunnel tunnel, TunnelFactory tf) {
        return new BlacklistTest(clientA, clientB, tunnel, tf);
    }
    
    private Test createWhitelistTest(Client clientA, Client clientB, Tunnel tunnel, TunnelFactory tf) {
        return new WhitelistTest(clientA, clientB, tunnel, tf);
    }
    
    private boolean _stopOnFailure = true;
    private static int _tunnelIndex = 0;
    private static TunnelFactory _tunnelFactory = null;
}
